<!DOCTYPE html>
<html>
    <head>
        <title>Test</title>
        {{include_meta}}
        {{include_helpers}}
        {{include_static 'style.css'}}
    </head>
    <body>
        <div class="page">
            <a href="/" class="title"><h1>Test</h1></a>
            <div class="nav">
                <a href="/">{{=current.T("partly cloudy")}}</a>
            </div>
            {{block main}}
            {{include}}
            {{end}}
        </div>
    </body>
</html>
