# -*- coding: utf-8 -*-
"""
    emmett.__main__
    ----------------

    Alias for Emmett CLI.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

if __name__ == "__main__":
    from .cli import main

    main(as_module=True)
